# Copyright 2019 kubeflow.org.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import json
from abc import ABC, abstractmethod
from typing import Any, List, Mapping, Optional, TypeVar, Union

import ml_metadata
from ml_metadata.metadata_store import metadata_store
from ml_metadata.proto import metadata_store_pb2 as mlpb
from retrying import retry
"""This module contains the Python API for logging metadata of machine learning
workflows to the Kubeflow Metadata service.
"""

_WORKSPACE_PROPERTY_NAME = '__kf_workspace__'
_RUN_PROPERTY_NAME = '__kf_run__'
_ALL_META_PROPERTY_NAME = '__ALL_META__'


class Artifact(ABC):

  @property
  @classmethod
  @abstractmethod
  def ARTIFACT_TYPE_NAME(cls) -> str:
    pass

  @abstractmethod
  def serialization(self) -> mlpb.Artifact:
    pass

  @staticmethod
  def is_duplicated(a: mlpb.Artifact, b: mlpb.Artifact):
    '''Checks if two artifacts are duplicated.

    The artifacts may be considered duplication even if not all the fields are
    the same as in mlpb.Artifact. For example, two models can be considered the
    same if they have the same uri, name and version.

    Returns:
      True or False for duplication.
    '''
    return False


class Store(object):
  """Metadata Store that connects to the Metadata gRPC service."""

  def __init__(self,
               grpc_host: str = "metadata-grpc-service.kubeflow",
               grpc_port: int = 8080,
               root_certificates: Optional[bytes] = None,
               private_key: Optional[bytes] = None,
               certificate_chain: Optional[bytes] = None):
    """
    Args:
      grpc_host: Required gRPC service host, e.g."metadata-grpc-service.kubeflow".
      grpc_host: Required gRPC service port.
      root_certificates: Optional SSL certificate for secure connection.
      private_key: Optional private_key for secure connection.
      certificate_chain: Optional certificate_chain for secure connection.

    The optional parameters are the same as in grpc.ssl_channel_credentials.
    https://grpc.github.io/grpc/python/grpc.html#grpc.ssl_channel_credentials
    """
    config = mlpb.MetadataStoreClientConfig()
    config.host = grpc_host
    config.port = grpc_port
    if private_key:
      config.ssl_config.client_key = private_key
    if root_certificates:
      config.ssl_config.custom_ca = root_certificates
    if certificate_chain:
      config.ssl_config.server_cert = certificate_chain

    self.store = metadata_store.MetadataStore(config)


class Workspace(object):
  """Groups a set of runs of pipelines, notebooks and their related artifacts
  and executions.
  """
  CONTEXT_TYPE_NAME = "kubeflow.org/alpha/workspace"

  def __init__(self,
               store: Store = None,
               name: str = None,
               description: Optional[str] = None,
               labels: Optional[Mapping[str, str]] = None,
               reuse_workspace_if_exists: Optional[bool] = True,
               backend_url_prefix: Optional[str] = None):
    """
    Args:
      store: Required store object to connect to MLMD gRPC service.
      name: Required name for the workspace.
      description: Optional string for description of the workspace.
      labels: Optional key/value string pairs to label the workspace.
      reuse_workspace_if_exists: Optional boolean value to indicate whether a
        workspace of the same name should be reused.
      backend_url_prefix: Deprecated. Please use 'store' parameter.

    Raises:
      ValueError: If a workspace of the same name already exists and
      `reuse_workspace_if_exists` is set to False.
    """
    if backend_url_prefix:
      raise ValueError(
          "'backend_url_prefix' is deprecated. Please set Metadata.Store "
          "parameter to connect to the metadata gRPC service.")
    if name is None or type(name) != str:
      raise ValueError("'name' must be set and in string type.")
    if not store or type(store) != Store:
      raise ValueError("'store' must be set as metadata.Store")
    self.store = store.store
    self.name = name
    self.description = description
    self.labels = labels
    self.context_id = self._get_context_id(reuse_workspace_if_exists)

  def list(self, artifact_type_name: str = None) -> List[Artifact]:
    """List all artifacts of a given type.

    Args:
      artifact_type_name {str} name of artifact type.
      Returns a list of artifacts of the provided typed associated
      with this workspace. Each artifact is represented as a dict.

    Returns:
      A list of Artifact objects.
    """
    if artifact_type_name is None:
      artifact_type_name = Model.ARTIFACT_TYPE_NAME
    response = _retry(
        lambda: self.store.get_artifacts_by_type(artifact_type_name))
    results = []
    for artifact in response:
      flat = self._flat(artifact)
      if "workspace" in flat and flat["workspace"] == self.name:
        results.append(flat)
    return results

  def _flat(self, artifact):
    if artifact is None:
      raise ValueError("'artifact' must be set.")
    result = {
        "id": artifact.id,
    }
    if artifact.custom_properties is not None:
      if _WORKSPACE_PROPERTY_NAME in artifact.custom_properties:
        result["workspace"] = artifact.custom_properties[
            _WORKSPACE_PROPERTY_NAME].string_value
      if _RUN_PROPERTY_NAME in artifact.custom_properties:
        result["run"] = artifact.custom_properties[
            _RUN_PROPERTY_NAME].string_value
    if not artifact.properties:
      return result
    for k, v in artifact.properties.items():
      if k != _ALL_META_PROPERTY_NAME:
        if v.string_value is not None:
          result[k] = v.string_value
        elif v.int_value is not None:
          result[k] = v.int_value
        else:
          result[k] = v.double_value
    if not _ALL_META_PROPERTY_NAME in artifact.properties:
      return result
    # Pick up all nested objects stored in the __ALL_META__ field.
    all_meta = artifact.properties[_ALL_META_PROPERTY_NAME].string_value
    for k, v in json.loads(all_meta).items():
      if not k in result:
        result[k] = v
    return result

  def _get_context_id(self, reuse_workspace_if_exists):
    ctx = self._get_existing_context()
    if ctx is not None:
      if reuse_workspace_if_exists:
        return ctx.id
      else:
        raise ValueError(
            'Workspace name {} already exists with id {}. You can initialize workspace with reuse_workspace_if_exists=True if want to reuse it'
            .format(self.name, ctx.id))
    # Create new context type or get the existing type id.
    ctx_type = mlpb.ContextType(name=self.CONTEXT_TYPE_NAME,
                                properties={
                                    "description": mlpb.STRING,
                                    "labels": mlpb.STRING
                                })
    ctx_type_id = _retry(lambda: self.store.put_context_type(ctx_type))

    # Add new context for workspace.
    prop = {}
    if self.description is not None:
      prop["description"] = mlpb.Value(string_value=self.description)
    if self.labels is not None:
      prop["labels"] = mlpb.Value(string_value=json.dumps(self.labels))
    ctx = mlpb.Context(
        type_id=ctx_type_id,
        name=self.name,
        properties=prop,
    )
    ctx_id = _retry(lambda: self.store.put_contexts([ctx])[0])
    return ctx_id

  def _get_existing_context(self):
    contexts = _retry(
        lambda: self.store.get_contexts_by_type(self.CONTEXT_TYPE_NAME))
    for ctx in contexts:
      if ctx.name == self.name:
        return ctx
    return None


class Run(object):
  """Run captures a run of pipeline or notebooks in a workspace and group
  executions.
  """

  def __init__(self,
               workspace: Workspace = None,
               name: str = None,
               description: Optional[str] = None):
    """
    Args:
      workspace: Required workspace object to which this run belongs.
      name: Required name of this run.
      description: Optional description.
    """
    if workspace is None:
      raise ValueError("'workspace' must be set.")
    if name is None or type(name) != str:
      raise ValueError("'name' must be set and in string type.")
    self.workspace = workspace
    self.name = name
    self.description = description


class Execution(object):
  """Captures a run of pipeline or notebooks in a workspace and group executions.

  Execution also serves as object for logging artifacts as its input or output.
  """
  EXECUTION_TYPE_NAME = "kubeflow.org/alpha/execution"

  def __init__(self,
               name: str = None,
               workspace: Workspace = None,
               run: Optional[Run] = None,
               description: Optional[str] = None):
    """
    Args:
      name: Required name of this run.
      workspace: Required workspace object where this execution belongs to.
      run: Optional run object.
      description: Optional description.

    Creates a new execution in a workspace and run.
    The execution.log_XXX() methods will attach corresponding artifacts as the
    input or output of this execution.
    """
    if workspace is None:
      raise ValueError("'workspace' must be set.")
    if name is None or type(name) != str:
      raise ValueError("'name' must be set and in string type.")
    self.id = None
    self.name = name
    self.workspace = workspace
    self.run = run
    self.description = description
    self.create_time = _get_rfc3339_time()
    self._type_id = _retry(lambda: self.workspace.store.get_execution_type(
        Execution.EXECUTION_TYPE_NAME).id)
    self.id = _retry(
        lambda: self.workspace.store.put_executions([self.serialized()])[0])
    _retry(lambda: self.workspace.store.put_attributions_and_associations([], [
        mlpb.Association(context_id=self.workspace.context_id,
                         execution_id=self.id)
    ]))

  def __repr__(self):
    field_names = self.__dict__.keys()
    fields_str = ", ".join(
        "{}={!r}".format(name, getattr(self, name)) for name in field_names)
    return "{0.__class__.__module__}.{0.__class__.__qualname__}({1})".format(
        self, fields_str)

  def serialized(self):
    properties = {
        "name": mlpb.Value(string_value=self.name),
        "create_time": mlpb.Value(string_value=self.create_time),
        "description": mlpb.Value(string_value=self.description),
    }
    _del_none_properties(properties)

    custom_properties = {}
    if self.workspace is not None:
      custom_properties[_WORKSPACE_PROPERTY_NAME] = mlpb.Value(
          string_value=self.workspace.name)
    if self.run is not None:
      custom_properties[_RUN_PROPERTY_NAME] = mlpb.Value(
          string_value=self.run.name)
    return mlpb.Execution(type_id=self._type_id,
                          properties=properties,
                          custom_properties=custom_properties)

  def log_input(self, artifact: Artifact) -> Artifact:
    """ Log an artifact as an input of this execution.

    Args:
      artifact: Model, DataSet, Metrics or customized artifact type.

    Returns:
      The same artifact with artifact.id set.
    """
    if artifact is None:
      raise ValueError("'artifact' must be set.")
    self._log(artifact)
    input_event = mlpb.Event(artifact_id=artifact.id,
                             execution_id=self.id,
                             type=mlpb.Event.INPUT)
    _retry(lambda: self.workspace.store.put_events([input_event]))
    return artifact

  def log_output(self, artifact: Artifact) -> Artifact:
    """ Log an artifact as an input of this execution.

    Args:
      artifact: Model, DataSet, Metrics or customized artifact type.

    Returns:
      The same artifact with artifact.id set.
    """
    if artifact is None:
      raise ValueError("'artifact' must be set.")
    self._log(artifact)
    output_event = mlpb.Event(artifact_id=artifact.id,
                              execution_id=self.id,
                              type=mlpb.Event.OUTPUT)
    _retry(lambda: self.workspace.store.put_events([output_event]))
    return artifact

  def _log(self, artifact):
    """Log artifact into metadata store."""
    # Sanity checks for artifact.
    if artifact is None:
      raise ValueError("'artifact' must be set.")
    try:
      type_id = _retry(lambda: self.workspace.store.get_artifact_type(
          artifact.ARTIFACT_TYPE_NAME).id)
    except Exception as e:
      raise ValueError("invalid artifact type %s: exception %s",
                       artifact.ARTIFACT_TYPE_NAME, e)

    # if id is set, then this artifact is already saved in database.
    if artifact.id is not None:
      self._check_artifact_id(artifact.id)
      return artifact

    # Construct a new artifact serialization.
    ser = artifact.serialization()
    ser.type_id = type_id
    if _WORKSPACE_PROPERTY_NAME in ser.custom_properties:
      raise ValueError("custom_properties contains reserved key %s" %
                       _WORKSPACE_PROPERTY_NAME)
    if _RUN_PROPERTY_NAME in ser.custom_properties:
      raise ValueError("custom_properties contains reserved key %s" %
                       _RUN_PROPERTY_NAME)
    if self.workspace is not None:
      ser.custom_properties[
          _WORKSPACE_PROPERTY_NAME].string_value = self.workspace.name
    if self.run is not None:
      ser.custom_properties[_RUN_PROPERTY_NAME].string_value = self.run.name

    # Deduplicate artifact for existing one in the database.
    pbs = _retry(
        lambda: self.workspace.store.get_artifacts_by_uri(artifact.uri))
    for pb in pbs:
      if artifact.is_duplicated(ser, pb):
        artifact.id = pb.id
        return artifact

    artifact.id = _retry(lambda: self.workspace.store.put_artifacts([ser])[0])
    _retry(lambda: self.workspace.store.put_attributions_and_associations([
        mlpb.Attribution(context_id=self.workspace.context_id,
                         artifact_id=artifact.id)
    ], []))
    return artifact

  def _check_artifact_id(self, aid):
    try:
      pbs = _retry(lambda: self.workspace.store.get_artifacts_by_id([aid]))
    except Exception as e:
      raise ValueError("invalid artifact id {}: {}".format(aid, e))
    if len(pbs) != 1:
      raise ValueError(
          "invalid artifact id {}: artifacts with this id: {}".format(aid, pbs))


class DataSet(Artifact):
  """ Dataset captures a data set in a machine learning workflow.

  Attributes:

    uri: Required uri of the data set.
    name: Required name of the data set.
    workspace: Optional name of the workspace.
    description: Optional description of the data set.
    owner: Optional owner of the data set.
    version: Optional version tagged by the user.
    query: Optional query string on how this data set being fetched from a data
      source.
    labels: Optional string key value pairs for labels.

  Example:
      >>> metadata.DataSet(description="an example data",
      ...                  name="mytable-dump",
      ...                  owner="owner@my-company.org",
      ...                  uri="file://path/to/dataset",
      ...                  version="v1.0.0",
      ...                  query="SELECT * FROM mytable",
      ...                  labels={"label1","val1"}))
  """
  ARTIFACT_TYPE_NAME = "kubeflow.org/alpha/data_set"

  def __init__(self,
               uri: str = None,
               name: str = None,
               workspace: Optional[str] = None,
               description: Optional[str] = None,
               owner: Optional[str] = None,
               version: Optional[str] = None,
               query: Optional[str] = None,
               labels: Optional[Mapping[str, str]] = None,
               **kwargs):
    """
    Args:
      uri: Required uri of the data set.
      name: Required name of the data set.
      workspace: Optional name of the workspace.
      description: Optional description of the data set.
      owner: Optional owner of the data set.
      version: Optional version tagged by the user.
      query: Optional query string on how this data set being fetched from
        a data source.
      labels: Optional string key value pairs for labels.
      kwargs: Optional additional keyword arguments are saved as additional
        properties of this dataset.
    """
    if uri is None or type(uri) != str:
      raise ValueError("'uri' must be set and in string type.")
    if name is None or type(name) != str:
      raise ValueError("'name' must be set and in string type.")
    self.workspace = workspace
    self.name = name
    self.description = description
    self.owner = owner
    self.uri = uri
    self.version = version
    self.query = query
    self.labels = labels
    self.id = None
    self.create_time = _get_rfc3339_time()
    self.kwargs = kwargs

  def __repr__(self):
    field_names = self.__dict__.keys()
    fields_str = ", ".join(
        "{}={!r}".format(name, getattr(self, name)) for name in field_names)
    return "{0.__class__.__module__}.{0.__class__.__qualname__}({1})".format(
        self, fields_str)

  def serialization(self):
    data_set_artifact = mlpb.Artifact(
        uri=self.uri,
        properties={
            "name":
                mlpb.Value(string_value=self.name),
            "create_time":
                mlpb.Value(string_value=self.create_time),
            "description":
                mlpb.Value(string_value=self.description),
            "query":
                mlpb.Value(string_value=self.query),
            "version":
                mlpb.Value(string_value=self.version),
            "owner":
                mlpb.Value(string_value=self.owner),
            _ALL_META_PROPERTY_NAME:
                mlpb.Value(string_value=json.dumps(self.__dict__)),
        })
    _del_none_properties(data_set_artifact.properties)
    return data_set_artifact

  @staticmethod
  def is_duplicated(a, b):
    ap = a.properties
    bp = b.properties
    aws = a.custom_properties.get(_WORKSPACE_PROPERTY_NAME)
    bws = b.custom_properties.get(_WORKSPACE_PROPERTY_NAME)
    return a.type_id == b.type_id and a.uri == b.uri and ap.get(
        "name") == bp.get("name") and ap.get("version") == bp.get(
            "version") and aws == bws


class Model(Artifact):
  """Captures a machine learning model.

  Attributes:
      uri: Required uri of the metrics.
      name: Required name of the metrics.
      workspace: Optional name of the workspace.
      description: Optional description of the metrics.
      owner: Optional owner of the metrics.
      model_type: Optional type of the model.
      training_framework: Optional framework used to train the model.
      hyperparameters: Optional map from hyper param name to its value.
      labels: Optional string key value pairs for labels.
      kwargs: Optional additional keyword arguments are saved as additional
        properties of this model.

  Example:
      >>> metadata.Model(name="MNIST",
      ...                description="model to recognize handwritten digits",
      ...                owner="someone@kubeflow.org",
      ...                uri="gcs://my-bucket/mnist",
      ...                model_type="neural network",
      ...                training_framework={
      ...                   "name": "tensorflow",
      ...                   "version": "v1.0"
      ...                },
      ...                hyperparameters={
      ...                   "learning_rate": 0.5,
      ...                   "layers": [10, 3, 1],
      ...                   "early_stop": True
      ...                },
      ...                version="v0.0.1",
      ...                labels={"mylabel": "l1"}))
  """

  ARTIFACT_TYPE_NAME = "kubeflow.org/alpha/model"

  def __init__(self,
               uri: str = None,
               name: str = None,
               workspace: Optional[str] = None,
               description: Optional[str] = None,
               owner: Optional[str] = None,
               version: Optional[str] = None,
               model_type: Optional[str] = None,
               training_framework: Optional[Any] = None,
               hyperparameters: Optional[Mapping[str, float]] = None,
               labels: Optional[Mapping[str, str]] = None,
               **kwargs):
    """
    Args:
      uri: Required uri of the metrics.
      name: Required name of the metrics.
      workspace: Optional name of the workspace.
      description: Optional description of the metrics.
      owner: Optional owner of the metrics.
      model_type: Optional type of the model.
      training_framework: Optional framework used to train the model.
      hyperparameters: Optional map from hyper param name to its value.
      labels: Optional string key value pairs for labels.
      kwargs: Optional additional keyword arguments are saved as additional
        properties of this model.
    """
    if uri is None or type(uri) != str:
      raise ValueError("'uri' must be set and in string type.")
    if name is None or type(name) != str:
      raise ValueError("'name' must be set and in string type.")
    self.workspace = workspace
    self.name = name
    self.description = description
    self.owner = owner
    self.uri = uri
    self.version = version
    self.model_type = model_type
    self.training_framework = training_framework
    self.hyperparameters = hyperparameters
    self.labels = labels
    self.id = None
    self.create_time = _get_rfc3339_time()
    self.kwargs = kwargs

  def __repr__(self):
    field_names = self.__dict__.keys()
    fields_str = ", ".join(
        "{}={!r}".format(name, getattr(self, name)) for name in field_names)
    return "{0.__class__.__module__}.{0.__class__.__qualname__}({1})".format(
        self, fields_str)

  def serialization(self):
    model_artifact = mlpb.Artifact(
        uri=self.uri,
        properties={
            "name":
                mlpb.Value(string_value=self.name),
            "create_time":
                mlpb.Value(string_value=self.create_time),
            "description":
                mlpb.Value(string_value=self.description),
            "model_type":
                mlpb.Value(string_value=self.model_type),
            "version":
                mlpb.Value(string_value=self.version),
            "owner":
                mlpb.Value(string_value=self.owner),
            _ALL_META_PROPERTY_NAME:
                mlpb.Value(string_value=json.dumps(self.__dict__)),
        })
    _del_none_properties(model_artifact.properties)
    return model_artifact

  @staticmethod
  def is_duplicated(a, b):
    ap = a.properties
    bp = b.properties
    aws = a.custom_properties.get(_WORKSPACE_PROPERTY_NAME)
    bws = b.custom_properties.get(_WORKSPACE_PROPERTY_NAME)
    return a.type_id == b.type_id and a.uri == b.uri and ap.get(
        "name") == bp.get("name") and ap.get("version") == bp.get(
            "version") and aws == bws


class Metrics(Artifact):
  """Captures an evaluation metrics of a model on a data set.

  Attributes:
    uri: Required uri of the metrics.
    name: Required name of the metrics.
    workspace: Optional name of the workspace.
    description: Optional description of the metrics.
    owner: Optional owner of the metrics.
    data_set_id: Optional id of the data set used for evaluation.
    model_id: Optional id of a evaluated model.
    metrics_type: Optional type of the evaluation.
    values: Optional map from metrics name to its value.
    labels: Optional string key value pairs for labels.

  Example:
    >>> metadata.Metrics(
    ...        name="MNIST-evaluation",
    ...        description=
    ...        "validating the MNIST model to recognize handwritten digits",
    ...        owner="someone@kubeflow.org",
    ...        uri="gcs://my-bucket/mnist-eval.csv",
    ...        data_set_id="123",
    ...        model_id="12345",
    ...        metrics_type=metadata.Metrics.VALIDATION,
    ...        values={"accuracy": 0.95},
    ...        labels={"mylabel": "l1"}))
  """

  ARTIFACT_TYPE_NAME = "kubeflow.org/alpha/metrics"

  # Possible evaluation metrics types.
  TRAINING = "training"
  VALIDATION = "validation"
  TESTING = "testing"
  PRODUCTION = "production"

  def __init__(self,
               uri: str = None,
               name: str = None,
               workspace: Optional[str] = None,
               description: Optional[str] = None,
               owner: Optional[str] = None,
               data_set_id: Optional[str] = None,
               model_id: Optional[str] = None,
               metrics_type: Optional[str] = None,
               values: Optional[Mapping[str, float]] = None,
               labels: Optional[Mapping[str, str]] = None,
               **kwargs):
    """
    Args:
      uri: Required uri of the metrics.
      name: Required name of the metrics.
      workspace: Optional name of the workspace.
      description: Optional description of the metrics.
      owner: Optional owner of the metrics.
      data_set_id: Optional id of the data set used for evaluation.
      model_id: Optional id of a evaluated model.
      metrics_type: Optional type of the evaluation.
      values: Optional map from metrics name to its value.
      labels: Optional string key value pairs for labels.
      kwargs: additional keyword arguments are saved as additional properties
        of this metrics.
    """
    if uri is None or type(uri) != str:
      raise ValueError("'uri' must be set and in string type.")
    if name is None or type(name) != str:
      raise ValueError("'name' must be set and in string type.")
    self.workspace = workspace
    self.name = name
    self.description = description
    self.owner = owner
    self.uri = uri
    self.data_set_id = data_set_id
    self.model_id = model_id
    self.metrics_type = metrics_type
    self.values = values
    self.labels = labels
    self.id = None
    self.create_time = _get_rfc3339_time()
    self.kwargs = kwargs

  def __repr__(self):
    field_names = self.__dict__.keys()
    fields_str = ", ".join(
        "{}={!r}".format(name, getattr(self, name)) for name in field_names)
    return "{0.__class__.__module__}.{0.__class__.__qualname__}({1})".format(
        self, fields_str)

  def serialization(self):
    metrics_artifact = mlpb.Artifact(
        uri=self.uri,
        properties={
            "name":
                mlpb.Value(string_value=self.name),
            "create_time":
                mlpb.Value(string_value=self.create_time),
            "description":
                mlpb.Value(string_value=self.description),
            "metrics_type":
                mlpb.Value(string_value=self.metrics_type),
            "data_set_id":
                mlpb.Value(string_value=self.data_set_id),
            "model_id":
                mlpb.Value(string_value=self.model_id),
            "owner":
                mlpb.Value(string_value=self.owner),
            _ALL_META_PROPERTY_NAME:
                mlpb.Value(string_value=json.dumps(self.__dict__)),
        })
    _del_none_properties(metrics_artifact.properties)
    return metrics_artifact


@retry(wait_exponential_multiplier=500, stop_max_delay=4000)
def _retry(f):
  '''retry function f with exponential backoff'''
  return f()


def _get_rfc3339_time():
  return datetime.datetime.utcnow().isoformat("T") + "Z"


def _del_none_properties(dict):
  keys = list(dict.keys())
  for k in keys:
    if not any((dict[k].string_value, dict[k].int_value, dict[k].double_value)):
      del dict[k]



import unittest
import uuid
from unittest.mock import patch

import ml_metadata
import pytest
from kubeflow.metadata import metadata
from ml_metadata.proto import metadata_store_pb2 as mlpb

GRPC_HOST = "127.0.0.1"
GRPC_PORT = 8081


class TestMetedata(unittest.TestCase):

  def test_log_metadata_successfully(self):
    store = metadata.Store(grpc_host=GRPC_HOST, grpc_port=GRPC_PORT)
    ws_name = "test_log_metadata_successfully_ws_" + str(uuid.uuid4())
    ws1 = metadata.Workspace(store=store,
                             name=ws_name,
                             description="a workspace for testing",
                             labels={"n1": "v1"})

    r = metadata.Run(
        workspace=ws1,
        name="first run",
        description="first run in ws_1",
    )

    trainer = metadata.Execution(
        name="test execution",
        workspace=ws1,
        run=r,
        description="an execution",
    )
    self.assertIsNotNone(trainer.id)

    data_set = trainer.log_input(
        metadata.DataSet(description="an example data",
                         name="mytable-dump",
                         owner="owner@my-company.org",
                         uri="file://path/to/dataset",
                         version=str(uuid.uuid4()),
                         query="SELECT * FROM mytable"))
    self.assertIsNotNone(data_set.id)
    self.assertIsNotNone(repr(data_set))

    metrics = trainer.log_output(
        metadata.Metrics(
            name="MNIST-evaluation",
            description=
            "validating the MNIST model to recognize handwritten digits",
            owner="someone@kubeflow.org",
            uri="gcs://my-bucket/mnist-eval.csv",
            data_set_id="123",
            model_id="12345",
            metrics_type=metadata.Metrics.VALIDATION,
            values={"accuracy": 0.95},
            labels={"mylabel": "l1"}))
    self.assertIsNotNone(metrics.id)
    self.assertIsNotNone(repr(metrics))
    print(repr(metrics))

    model_version = str(uuid.uuid4())
    model = trainer.log_output(
        metadata.Model(name="MNIST",
                       description="model to recognize handwritten digits",
                       owner="someone@kubeflow.org",
                       uri="gcs://my-bucket/mnist",
                       model_type="neural network",
                       training_framework={
                           "name": "tensorflow",
                           "version": "v1.0"
                       },
                       hyperparameters={
                           "learning_rate": 0.5,
                           "layers": [10, 3, 1],
                           "early_stop": True
                       },
                       version=model_version,
                       labels={"mylabel": "l1"}))
    self.assertIsNotNone(model.id)
    self.assertIsNotNone(repr(model))

    serving_application = metadata.Execution(
        name="serving model",
        workspace=ws1,
        description="an execution to represent model serving component",
    )
    self.assertIsNotNone(serving_application.id)
    # Use model name, version, uri to uniquely identify existing model.
    served_model = metadata.Model(
        name="MNIST",
        uri="gcs://my-bucket/mnist",
        version=model_version,
    )
    serving_application.log_input(served_model)

    # Test listing artifacts in a workspace
    assert len(ws1.list()) > 0
    assert len(ws1.list(metadata.Model.ARTIFACT_TYPE_NAME)) > 0
    assert len(ws1.list(metadata.Metrics.ARTIFACT_TYPE_NAME)) > 0
    assert len(ws1.list(metadata.DataSet.ARTIFACT_TYPE_NAME)) > 0

    # Test lineage tracking.
    model_events = ws1.store.get_events_by_artifact_ids([model.id])
    self.assertEqual(len(model_events), 2)
    execution_ids = set(e.execution_id for e in model_events)
    assert execution_ids == set([serving_application.id, trainer.id])
    trainer_events = ws1.store.get_events_by_execution_ids([trainer.id])
    artifact_ids = set(e.artifact_id for e in trainer_events)
    assert artifact_ids == set([model.id, metrics.id, data_set.id])

    # Test context, attributions, associations creation.
    ctx = ws1._get_existing_context()
    assert ctx.id is not None
    artifacts = ws1.store.get_artifacts_by_context(ctx.id)
    assert set(a.id for a in artifacts) == set(
        [model.id, metrics.id, data_set.id])
    executions = ws1.store.get_executions_by_context(ctx.id)
    assert set(e.id for e in executions) == set(
        [serving_application.id, trainer.id])

  def test_log_invalid_artifacts_should_fail(self):
    store = metadata.Store(grpc_host=GRPC_HOST, grpc_port=GRPC_PORT)
    ws = metadata.Workspace(store=store,
                            name="ws_1",
                            description="a workspace for testing",
                            labels={"n1": "v1"})
    e = metadata.Execution(name="test execution", workspace=ws)
    artifact1 = ArtifactFixture(
        mlpb.Artifact(uri="gs://uri",
                      custom_properties={
                          metadata._WORKSPACE_PROPERTY_NAME:
                              mlpb.Value(string_value="ws1"),
                      }))
    self.assertRaises(ValueError, e.log_input, artifact1)
    artifact2 = ArtifactFixture(
        mlpb.Artifact(uri="gs://uri",
                      custom_properties={
                          metadata._RUN_PROPERTY_NAME:
                              mlpb.Value(string_value="run1"),
                      }))
    self.assertRaises(ValueError, e.log_output, artifact2)

  def test_log_metadata_successfully_with_minimum_information(self):
    store = metadata.Store(grpc_host=GRPC_HOST, grpc_port=GRPC_PORT)
    ws1 = metadata.Workspace(store=store, name="ws_1")
    r = metadata.Run(workspace=ws1, name="first run")
    e = metadata.Execution(name="test execution", workspace=ws1, run=r)
    self.assertIsNotNone(e.id)

    data_set = e.log_input(
        metadata.DataSet(name="mytable-dump", uri="file://path/to/dataset"))
    self.assertIsNotNone(data_set.id)
    data_set_id = data_set.id
    # ID should not change after logging twice.
    e.log_input(data_set)
    self.assertEqual(data_set_id, data_set.id)

    metrics = e.log_output(
        metadata.Metrics(name="MNIST-evaluation",
                         uri="gcs://my-bucket/mnist-eval.csv"))
    self.assertIsNotNone(metrics.id)
    metrics_id = metrics.id
    # ID should not change after logging twice.
    e.log_output(metrics)
    self.assertEqual(metrics_id, metrics.id)

    model = e.log_output(
        metadata.Model(name="MNIST", uri="gcs://my-bucket/mnist"))
    self.assertIsNotNone(model.id)
    model_id = model.id
    # ID should not change after logging twice.
    e.log_output(model)
    self.assertEqual(model_id, model.id)

  def test_invalid_workspace_should_fail(self):
    with self.assertRaises(ValueError):
      ws1 = metadata.Workspace(name="ws_1",
                               description="a workspace for testing",
                               labels={"n1": "v1"})

    with self.assertRaises(ValueError):
      ws1 = metadata.Workspace(store="127.1.0.1:8080",
                               name="ws_1",
                               description="a workspace for testing",
                               labels={"n1": "v1"})

  def test_creating_workspace_with_existing_name(self):
    store = metadata.Store(grpc_host=GRPC_HOST, grpc_port=GRPC_PORT)
    ws_name = "non_unique_ws_" + str(uuid.uuid4())
    ws1 = metadata.Workspace(store=store,
                             name=ws_name,
                             description="a workspace for testing",
                             labels={"n1": "v1"})

    ws2 = metadata.Workspace(store, ws_name)
    assert ws1.context_id == ws2.context_id

    with pytest.raises(ValueError, match=r".*exists with id.*"):
      metadata.Workspace(store, ws_name, reuse_workspace_if_exists=False)

  def test_artifact_deduplication(self):
    store = metadata.Store(grpc_host=GRPC_HOST, grpc_port=GRPC_PORT)
    ws1 = metadata.Workspace(store=store, name="workspace_one")
    ws2 = metadata.Workspace(store=store, name="workspace_two")
    r = metadata.Run(workspace=ws1, name="first run")
    e = metadata.Execution(name="test execution", workspace=ws1, run=r)
    e2 = metadata.Execution(name="execution 2", workspace=ws1)
    e3 = metadata.Execution(name="execution 3", workspace=ws2)
    self.assertIsNotNone(e.id)
    self.assertIsNotNone(e2.id)

    model = metadata.Model(name="MNIST",
                           uri="gcs://my-bucket/mnist",
                           model_type="neural network",
                           version="v0.0.1")
    model2 = metadata.Model(name="MNIST",
                            uri="gcs://my-bucket/mnist",
                            model_type="neural network",
                            version="v0.0.1")
    e.log_output(model)
    self.assertIsNotNone(model.id)
    e2.log_output(model2)
    self.assertIsNotNone(model2.id)
    self.assertEqual(model.id, model2.id)

  def test_is_duplicated_methods(self):
    for cls in [metadata.Model, metadata.DataSet]:
      m = mlpb_artifact(1, "gcs://123", "ws1", "name1", "v1")
      m1 = mlpb_artifact(1, "gcs://123", "ws1", "name1", "v1")
      self.assertTrue(cls.is_duplicated(m, m1))
      m2 = mlpb_artifact(1, "gcs://123", "ws1", "name1")
      self.assertFalse(cls.is_duplicated(m, m2))
      m3 = mlpb_artifact(1, "gcs://123", "ws1", "name2", "v1")
      self.assertFalse(cls.is_duplicated(m, m3))
      m4 = mlpb_artifact(1, "gcs://123", "ws2", "name1", "v1")
      self.assertFalse(cls.is_duplicated(m, m4))
      m5 = mlpb_artifact(1, "gcs://1234", "ws1", "name1", "v1")
      self.assertFalse(cls.is_duplicated(m, m5))
      m6 = mlpb_artifact(2, "gcs://123", "ws1", "name1", "v1")
      self.assertFalse(cls.is_duplicated(m, m6))


class CheckMetadataStore(object):

  def __init__(self, config, disable_upgrade_migration=None):
    assert disable_upgrade_migration is not None
    assert config.ssl_config is not None
    assert config.ssl_config.custom_ca is not None
    assert config.ssl_config.client_key is not None
    assert config.ssl_config.server_cert is not None


class ArtifactFixture(object):
  ARTIFACT_TYPE_NAME = "artifact_types/kubeflow.org/alpha/artifact_fixture"

  def __init__(self, serialization_fixture=None):
    self._fixture = serialization_fixture

  def serialization(self):
    return self._fixture


def mlpb_artifact(type_id, uri, workspace, name=None, version=None):
  properties = {}
  if name:
    properties["name"] = mlpb.Value(string_value=name)
  if version:
    properties["version"] = mlpb.Value(string_value=version)
  return mlpb.Artifact(uri=uri,
                       type_id=type_id,
                       properties=properties,
                       custom_properties={
                           metadata._WORKSPACE_PROPERTY_NAME:
                               mlpb.Value(string_value=workspace),
                       })


if __name__ == "__main__":
  unittest.main()



import tempfile
import logging
import os
import papermill

logger = logging.getLogger(__name__)

FILE_DIR = os.path.dirname(__file__)
NOTEBOOK_REL_PATH = "../sample/demo.ipynb"
NOTEBOOK_ABS_PATH = os.path.normpath(os.path.join(FILE_DIR, NOTEBOOK_REL_PATH))

GRPC_HOST = "127.0.0.1"
GRPC_PORT = 8081


def test_notebook():
  temp_dir = tempfile.mkdtemp()
  notebook_output_path = os.path.join(temp_dir, "out.ipynb")
  parameters = {
      "METADATA_STORE_HOST": GRPC_HOST,
      "METADATA_STORE_PORT": GRPC_PORT,
  }
  papermill.execute_notebook(NOTEBOOK_ABS_PATH,
                             notebook_output_path,
                             cwd=os.path.dirname(NOTEBOOK_ABS_PATH),
                             parameters=parameters,
                             log_output=True)
  check_notebook_output(notebook_output_path)


def check_notebook_output(output_path):
  num_cells = 0
  num_completed_cells = 0
  with open(output_path, 'r') as f:
    for lines in f:
      if lines.find('"status": "completed"') != -1:
        num_completed_cells = num_completed_cells + 1
      if lines.find('cell_type') != -1:
        num_cells = num_cells + 1
  with open(output_path, 'r') as f:
    assert num_cells == num_completed_cells, "Not all cells succeeded. Notebook output:\n {}".format(
        f.read())
